'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var react = require('react');
var scheduler = require('scheduler');
var useContextSelector = require('use-context-selector');

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _createForOfIteratorHelperLoose(o, allowArrayLike) {
  var it;

  if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) {
    if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") {
      if (it) o = it;
      var i = 0;
      return function () {
        if (i >= o.length) return {
          done: true
        };
        return {
          done: false,
          value: o[i++]
        };
      };
    }

    throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
  }

  it = o[Symbol.iterator]();
  return it.next.bind(it);
}

var isSSR = typeof window === 'undefined' || /ServerSideRendering/.test(window.navigator && window.navigator.userAgent);
var useIsoLayoutEffect = isSSR ? function (fn) {
  return fn();
} : react.useLayoutEffect;

var MAX_CHAIN_LENGTH = 10;
var SMALL_MAP_SIZE = 10;
var mCreate = function mCreate() {
  return [new Map()];
};
var mGet = function mGet(m, k) {
  for (var i = 0; i < m.length; ++i) {
    var v = m[i].get(k);

    if (v) {
      return v;
    }
  }
};

var squash = function squash(m) {
  var i = m.length - 1;
  var dst = new Map(m[i]);

  while (i) {
    --i;
    m[i].forEach(function (v, k) {
      dst.set(k, v);
    });
  }

  return [dst];
};

var mSet = function mSet(m, k, v) {
  if (m[0].size < SMALL_MAP_SIZE) {
    return [new Map(m[0]).set(k, v)].concat(m.slice(1));
  }

  if (m.length >= MAX_CHAIN_LENGTH) {
    m = squash(m);
  }

  return [new Map([[k, v]])].concat(m);
};
var mDel = function mDel(m, k) {
  if (m.length > 1) {
    m = squash(m);
  }

  var map = new Map(m[0]);
  map.delete(k);
  return [map];
};
var mKeys = function mKeys(m) {
  var keys = new Set();
  m.forEach(function (map) {
    for (var _iterator = _createForOfIteratorHelperLoose(map.keys()), _step; !(_step = _iterator()).done;) {
      var key = _step.value;
      keys.add(key);
    }
  });
  return keys;
};
var mMerge = function mMerge(m, newM) {
  m = [].concat(newM, m);

  if (m.length > MAX_CHAIN_LENGTH) {
    m = squash(m);
  }

  return m;
};
var mToPrintable = function mToPrintable(m, toPrintableK, toPrintableV) {
  if (m.length > 1) {
    m = squash(m);
  }

  return new Map([].concat(m[0].entries()).map(function (_ref) {
    var k = _ref[0],
        v = _ref[1];
    return [toPrintableK(k), toPrintableV(v)];
  }));
};

var warningObject = new Proxy({}, {
  get: function get() {
    throw new Error('Please use <Provider>');
  },
  apply: function apply() {
    throw new Error('Please use <Provider>');
  }
});

var addDependent = function addDependent(dependentsMap, atom, dependent) {
  var dependents = dependentsMap.get(atom);

  if (!dependents) {
    dependents = new Set();
    dependentsMap.set(atom, dependents);
  }

  dependents.add(dependent);
};

var deleteDependent = function deleteDependent(dependentsMap, dependent) {
  dependentsMap.forEach(function (dependents) {
    dependents.delete(dependent);
  });
};

var setDependencies = function setDependencies(dependentsMap, atom, dependencies) {
  deleteDependent(dependentsMap, atom);
  dependencies.forEach(function (dependency) {
    addDependent(dependentsMap, dependency, atom);
  });
};

var listDependents = function listDependents(dependentsMap, atom, excludeSelf) {
  var dependents = new Set(dependentsMap.get(atom));

  if (excludeSelf) {
    dependents.delete(atom);
  }

  return dependents;
};

var initialState = mCreate();

var getAtomStateValue = function getAtomStateValue(atom, state) {
  var atomState = mGet(state, atom);

  if (!atomState) {
    console.warn('Atom state not found. Possibly a bug. Please file an issue with repro.');
    return atom.init;
  }

  return atomState.value;
};

var readAtomState = function readAtomState(atom, prevState, setState, dependentsMap, force) {
  if (!force) {
    var atomState = mGet(prevState, atom);

    if (atomState) {
      return [atomState, prevState];
    }
  }

  var nextState = prevState;
  var error = undefined;
  var promise = undefined;
  var value = undefined;
  var dependencies = new Set();
  var isSync = true;

  try {
    var promiseOrValue = atom.read(function (a) {
      if (dependencies) {
        dependencies.add(a);
      } else {
        addDependent(dependentsMap, a, atom);
      }

      if (a !== atom) {
        var _readAtomState = readAtomState(a, nextState, setState, dependentsMap),
            _aState = _readAtomState[0],
            nextNextState = _readAtomState[1];

        if (isSync) {
          nextState = nextNextState;
        } else {
          // XXX is this really correct?
          setState(function (prev) {
            return mMerge(nextNextState, prev);
          });
        }

        if (_aState.readE) {
          throw _aState.readE;
        }

        if (_aState.readP) {
          throw _aState.readP;
        }

        return _aState.value;
      } // a === atom


      var aState = mGet(nextState, a);

      if (aState) {
        if (aState.readP) {
          throw aState.readP;
        }

        return aState.value;
      }

      return a.init; // this should not be undefined
    });

    if (promiseOrValue instanceof Promise) {
      promise = promiseOrValue.then(function (value) {
        setDependencies(dependentsMap, atom, dependencies);
        dependencies = null;
        setState(function (prev) {
          var _mGet;

          var prevPromise = (_mGet = mGet(prev, atom)) == null ? void 0 : _mGet.readP;

          if (prevPromise && prevPromise !== promise) {
            return prev;
          }

          return mSet(prev, atom, {
            value: value
          });
        });
      }).catch(function (e) {
        setState(function (prev) {
          var _mGet2;

          var prevPromise = (_mGet2 = mGet(prev, atom)) == null ? void 0 : _mGet2.readP;

          if (prevPromise && prevPromise !== promise) {
            return prev;
          }

          return mSet(prev, atom, {
            value: getAtomStateValue(atom, prev),
            readE: e instanceof Error ? e : new Error(e)
          });
        });
      });
    } else {
      setDependencies(dependentsMap, atom, dependencies);
      dependencies = null;
      value = promiseOrValue;
    }
  } catch (errorOrPromise) {
    if (errorOrPromise instanceof Promise) {
      promise = errorOrPromise;
    } else if (errorOrPromise instanceof Error) {
      error = errorOrPromise;
    } else {
      error = new Error(errorOrPromise);
    }
  }

  var nextAtomState = {
    readE: error,
    readP: promise,
    value: promise ? atom.init : value
  };
  nextState = mSet(nextState, atom, nextAtomState);
  isSync = false;
  return [nextAtomState, nextState];
};

var updateDependentsState = function updateDependentsState(atom, prevState, setState, dependentsMap) {
  var nextState = prevState;
  listDependents(dependentsMap, atom, true).forEach(function (dependent) {
    if (typeof dependent === 'symbol') return;

    var _readAtomState2 = readAtomState(dependent, nextState, setState, dependentsMap, true),
        dependentState = _readAtomState2[0],
        nextNextState = _readAtomState2[1];

    var promise = dependentState.readP;

    if (promise) {
      promise.then(function () {
        setState(function (prev) {
          return updateDependentsState(dependent, prev, setState, dependentsMap);
        });
      });
      nextState = nextNextState;
    } else {
      nextState = updateDependentsState(dependent, nextNextState, setState, dependentsMap);
    }
  });
  return nextState;
};

var readAtom = function readAtom(state, readingAtom, setState, dependentsMap, readPendingMap) {
  var prevState = readPendingMap.get(state) || state;

  var _readAtomState3 = readAtomState(readingAtom, prevState, setState, dependentsMap),
      atomState = _readAtomState3[0],
      nextState = _readAtomState3[1];

  if (nextState !== prevState) {
    readPendingMap.set(state, nextState);
  }

  return atomState;
};

var addAtom = function addAtom(id, atom, dependentsMap) {
  addDependent(dependentsMap, atom, id);
};

var delAtom = function delAtom(id, setGcCount, dependentsMap) {
  deleteDependent(dependentsMap, id);
  setGcCount(function (c) {
    return c + 1;
  }); // trigger re-render for gc
};

var gcAtom = function gcAtom(state, setState, dependentsMap) {
  var nextState = state;
  var deleted;

  do {
    deleted = false;
    mKeys(nextState).forEach(function (atom) {
      var _dependentsMap$get;

      var isEmpty = ((_dependentsMap$get = dependentsMap.get(atom)) == null ? void 0 : _dependentsMap$get.size) === 0;

      if (isEmpty) {
        nextState = mDel(nextState, atom);
        dependentsMap.delete(atom);
        deleted = true;
      }
    });
  } while (deleted);

  if (nextState !== state) {
    setState(nextState);
  }
};

var writeAtom = function writeAtom(writingAtom, update, setState, dependentsMap, addWriteThunk) {
  var pendingPromises = [];

  var writeAtomState = function writeAtomState(prevState, atom, update) {
    var prevAtomState = mGet(prevState, atom);

    if (prevAtomState && prevAtomState.writeP) {
      var promise = prevAtomState.writeP.then(function () {
        addWriteThunk(function (prev) {
          return writeAtomState(prev, atom, update);
        });
      });
      pendingPromises.push(promise);
      return prevState;
    }

    var nextState = prevState;
    var isSync = true;

    try {
      var promiseOrVoid = atom.write(function (a) {
        if (process.env.NODE_ENV !== 'production') {
          var s = mGet(nextState, a);

          if (s && s.readP) {
            // TODO will try to detect this
            console.warn('Reading pending atom state in write operation. We need to detect this and fallback. Please file an issue for repro.', a);
          }
        }

        return getAtomStateValue(a, nextState);
      }, function (a, v) {
        if (a === atom) {
          var nextAtomState = {
            value: v
          };

          if (isSync) {
            nextState = mSet(nextState, a, nextAtomState);
            nextState = updateDependentsState(a, nextState, setState, dependentsMap);
          } else {
            setState(function (prev) {
              return updateDependentsState(a, mSet(prev, a, nextAtomState), setState, dependentsMap);
            });
          }
        } else {
          if (isSync) {
            nextState = writeAtomState(nextState, a, v);
          } else {
            addWriteThunk(function (prev) {
              return writeAtomState(prev, a, v);
            });
          }
        }
      }, update);

      if (promiseOrVoid instanceof Promise) {
        pendingPromises.push(promiseOrVoid);

        var nextAtomState = _extends({}, mGet(nextState, atom), {
          writeP: promiseOrVoid.then(function () {
            addWriteThunk(function (prev) {
              return mSet(prev, atom, _extends({}, mGet(prev, atom), {
                writeP: undefined
              }));
            });
          })
        });

        nextState = mSet(nextState, atom, nextAtomState);
      }
    } catch (e) {
      if (pendingPromises.length) {
        pendingPromises.push(new Promise(function (_resolve, reject) {
          reject(e);
        }));
      } else {
        throw e;
      }
    }

    isSync = false;
    return nextState;
  };

  addWriteThunk(function (prevState) {
    return writeAtomState(prevState, writingAtom, update);
  });

  if (pendingPromises.length) {
    return new Promise(function (resolve, reject) {
      var loop = function loop() {
        var len = pendingPromises.length;

        if (len === 0) {
          resolve();
        } else {
          Promise.all(pendingPromises).then(function () {
            pendingPromises.splice(0, len);
            loop();
          }).catch(reject);
        }
      };

      loop();
    });
  }
};

var runWriteThunk = function runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdate, writeThunkQueue) {
  while (true) {
    if (!lastStateRef.current) {
      return;
    }

    if (writeThunkQueue.length === 0) {
      return;
    }

    var thunk = writeThunkQueue.shift();
    var prevState = pendingStateRef.current || lastStateRef.current;
    var nextState = thunk(prevState);

    if (nextState !== prevState) {
      pendingStateRef.current = nextState;
      scheduler.unstable_runWithPriority(scheduler.unstable_NormalPriority, function () {
        var pendingState = pendingStateRef.current;

        if (pendingState) {
          pendingStateRef.current = null;
          contextUpdate(function () {
            scheduler.unstable_runWithPriority(scheduler.unstable_UserBlockingPriority, function () {
              setState(pendingState);
            });
          });
        }
      });
    }
  }
};

var ActionsContext = useContextSelector.createContext(warningObject);
var StateContext = useContextSelector.createContext(warningObject);

var InnerProvider = function InnerProvider(_ref) {
  var r = _ref.r,
      children = _ref.children;
  var contextUpdate = useContextSelector.useContextUpdate(StateContext);

  if (!r.current) {
    r.current = contextUpdate;
  }

  return children;
};

var Provider = function Provider(_ref2) {
  var children = _ref2.children;
  var contextUpdateRef = react.useRef();
  var dependentsMapRef = react.useRef();

  if (!dependentsMapRef.current) {
    dependentsMapRef.current = new Map();
  }

  var readPendingMapRef = react.useRef();

  if (!readPendingMapRef.current) {
    readPendingMapRef.current = new WeakMap();
  }

  var _useState = react.useState(initialState),
      state = _useState[0],
      setStateOrig = _useState[1];

  var lastStateRef = react.useRef(null);
  var pendingStateRef = react.useRef(null);

  var setState = function setState(setStateAction) {
    lastStateRef.current = null;
    var pendingState = pendingStateRef.current;

    if (pendingState) {
      pendingStateRef.current = null;
      setStateOrig(pendingState);
    }

    setStateOrig(setStateAction);
  };

  useIsoLayoutEffect(function () {
    var readPendingMap = readPendingMapRef.current;
    var pendingState = readPendingMap.get(state);

    if (pendingState) {
      setState(pendingState);
      return;
    }

    lastStateRef.current = state;
  });

  var _useState2 = react.useState(0),
      gcCount = _useState2[0],
      setGcCount = _useState2[1]; // to trigger gc


  react.useEffect(function () {
    gcAtom(state, setState, dependentsMapRef.current);
  }, [state, gcCount]);
  var writeThunkQueueRef = react.useRef([]);
  react.useEffect(function () {
    runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdateRef.current, writeThunkQueueRef.current);
  }, [state]);
  var actions = react.useMemo(function () {
    return {
      add: function add(id, atom) {
        return addAtom(id, atom, dependentsMapRef.current);
      },
      del: function del(id) {
        return delAtom(id, setGcCount, dependentsMapRef.current);
      },
      read: function read(state, atom) {
        return readAtom(state, atom, setState, dependentsMapRef.current, readPendingMapRef.current);
      },
      write: function write(atom, update) {
        return writeAtom(atom, update, setState, dependentsMapRef.current, function (thunk) {
          writeThunkQueueRef.current.push(thunk);
          runWriteThunk(lastStateRef, pendingStateRef, setState, contextUpdateRef.current, writeThunkQueueRef.current);
        });
      }
    };
  }, []);

  if (process.env.NODE_ENV !== 'production') {
    // eslint-disable-next-line react-hooks/rules-of-hooks
    useDebugState(state);
  }

  return react.createElement(ActionsContext.Provider, {
    value: actions
  }, react.createElement(StateContext.Provider, {
    value: state
  }, react.createElement(InnerProvider, {
    r: contextUpdateRef
  }, children)));
};

var useDebugState = function useDebugState(state) {
  react.useDebugValue(state, function (s) {
    return mToPrintable(s, function (k) {
      var _k$debugLabel;

      return k.key + ":" + ((_k$debugLabel = k.debugLabel) != null ? _k$debugLabel : '<no debugLabel>');
    }, function (v) {
      return v.readE || v.readP || v.writeP || v.value;
    });
  });
};

var keyCount = 0; // global key count for all atoms

function atom(read, write) {
  var config = {
    key: ++keyCount
  };

  if (typeof read === 'function') {
    config.read = read;
  } else {
    config.init = read;

    config.read = function (get) {
      return get(config);
    };

    config.write = function (get, set, update) {
      set(config, typeof update === 'function' ? update(get(config)) : update);
    };
  }

  if (write) {
    config.write = write;
  }

  return config;
}

var isWritable = function isWritable(atom) {
  return !!atom.write;
};

function useAtom(atom) {
  var actions = useContextSelector.useContext(ActionsContext);
  var value = useContextSelector.useContextSelector(StateContext, react.useCallback(function (state) {
    var atomState = actions.read(state, atom);

    if (atomState.readE) {
      throw atomState.readE; // read error
    }

    if (atomState.readP) {
      throw atomState.readP; // read promise
    }

    if (atomState.writeP) {
      throw atomState.writeP; // write promise
    }

    return atomState.value;
  }, [atom, actions]));
  useIsoLayoutEffect(function () {
    var id = Symbol();
    actions.add(id, atom);
    return function () {
      actions.del(id);
    };
  }, [actions, atom]);
  var setAtom = react.useCallback(function (update) {
    if (isWritable(atom)) {
      return actions.write(atom, update);
    } else {
      throw new Error('not writable atom');
    }
  }, [atom, actions]);
  react.useDebugValue(value);
  return [value, setAtom];
}

var useBridge = function useBridge() {
  var actions = useContextSelector.useBridgeValue(ActionsContext);
  var state = useContextSelector.useBridgeValue(StateContext);
  return react.useMemo(function () {
    return [actions, state];
  }, [actions, state]);
};
var Bridge = function Bridge(_ref) {
  var value = _ref.value,
      children = _ref.children;
  var actions = value[0],
      state = value[1];
  return react.createElement(useContextSelector.BridgeProvider, {
    context: ActionsContext,
    value: actions
  }, react.createElement(useContextSelector.BridgeProvider, {
    context: StateContext,
    value: state
  }, children));
};

exports.Bridge = Bridge;
exports.Provider = Provider;
exports.atom = atom;
exports.useAtom = useAtom;
exports.useBridge = useBridge;
